#include "stdafx.h"
#include "ping.h"


class CMyApp : public CWinApp
{
protected:
  BOOL InitInstance() override;
  static CString AddressToString(const SOCKADDR* pSockAddr, int nSockAddrLen, int nFlags, UINT* pnSocketPort);
  static CString GetErrorMessage(DWORD dwError);
  static CString RTTAsString(DWORD dwRTT);
  static CString GetIpErrorString(IP_STATUS dwError);
};

#pragma warning(suppress: 26426)
CMyApp theApp;

class CMyCommandLineInfo : public CCommandLineInfo
{
public:
  CMyCommandLineInfo() noexcept;
  void ParseParam(const TCHAR* pszParam, BOOL bFlag, BOOL bLast) override;

//Our specific command line options
  CString m_sHostToResolve;
  CString m_sLocalBoundAddress;
  bool m_bResolveAddressesToHostnames;
  bool m_bPingTillStopped;
  int m_nRequestsToSend;
  UCHAR m_nTTL;
  UCHAR m_nTOS;
  WORD m_wDataRequestSize;
  DWORD m_dwTimeout;
  bool m_bDontFragment;
  bool m_bIPv6;
};

CMyCommandLineInfo::CMyCommandLineInfo() noexcept : m_bResolveAddressesToHostnames{false},
                                                    m_bPingTillStopped{false},
                                                    m_nRequestsToSend{1},
                                                    m_nTTL{128},
                                                    m_nTOS{0},
                                                    m_wDataRequestSize{32},
                                                    m_dwTimeout{5000},
                                                    m_bDontFragment{false},
                                                    m_bIPv6{false}
{
}

void CMyCommandLineInfo::ParseParam(LPCTSTR lpszParam, BOOL bFlag, BOOL bLast)
{
  if (bFlag)
  {
    //Form a CString version of the parameter (just makes it easier to parse with)
    CString sParam{lpszParam};

    if (_tcsicmp(lpszParam, _T("a")) == 0)
      m_bResolveAddressesToHostnames = true;
    else if (_tcsicmp(lpszParam, _T("t")) == 0)
      m_bPingTillStopped = true;
    else if (_tcsicmp(lpszParam, _T("f")) == 0)
      m_bDontFragment = true;
    else if (_tcsicmp(lpszParam, _T("4")) == 0)
      m_bIPv6 = false;
    else if (_tcsicmp(lpszParam, _T("6")) == 0)
      m_bIPv6 = true;
    else if ((sParam.Find(_T("n:")) == 0) && (sParam.GetLength() > 2))
      m_nRequestsToSend = _ttoi(sParam.Right(sParam.GetLength() - 2));
    else if ((sParam.Find(_T("i:")) == 0) && (sParam.GetLength() > 2))
#pragma warning(suppress: 26472)
      m_nTTL = static_cast<UCHAR>(_ttoi(sParam.Right(sParam.GetLength() - 2)));
    else if ((sParam.Find(_T("l:")) == 0) && (sParam.GetLength() > 2))
#pragma warning(suppress: 26472)
      m_wDataRequestSize = static_cast<WORD>(_ttoi(sParam.Right(sParam.GetLength() - 2)));
    else if ((sParam.Find(_T("v:")) == 0) && (sParam.GetLength() > 2))
#pragma warning(suppress: 26472)
      m_nTOS = static_cast<UCHAR>(_ttoi(sParam.Right(sParam.GetLength() - 2)));
    else if ((sParam.Find(_T("w:")) == 0) && (sParam.GetLength() > 2))
#pragma warning(suppress: 26472)
      m_dwTimeout = static_cast<DWORD>(_ttoi(sParam.Right(sParam.GetLength() - 2)));
    else if ((sParam.Find(_T("S:")) == 0) && (sParam.GetLength() > 2))
      m_sLocalBoundAddress = sParam.Right(sParam.GetLength() - 2);
    else
      CCommandLineInfo::ParseParam(lpszParam, bFlag, bLast);
  }
  else
    m_sHostToResolve = lpszParam;
}

CString CMyApp::AddressToString(const SOCKADDR* pSockAddr, int nSockAddrLen, int nFlags, UINT* pnSocketPort)
{
  //What will be the return value from this function
  CString sSocketAddress;

  int nResult{0};
#ifdef _UNICODE
  CStringW sName;
  nResult = GetNameInfoW(pSockAddr, nSockAddrLen, sName.GetBuffer(NI_MAXHOST), NI_MAXHOST, nullptr, 0, nFlags);
  sName.ReleaseBuffer();
#else
  CStringA sName;
  nResult = getnameinfo(pSockAddr, nSockAddrLen, sName.GetBuffer(NI_MAXHOST), NI_MAXHOST, nullptr, 0, nFlags);
  sName.ReleaseBuffer();
#endif
  if (nResult == 0)
  {
    sSocketAddress = sName;
    if (pnSocketPort != nullptr)
      *pnSocketPort = ntohs(SS_PORT(&pSockAddr));
  }

  return sSocketAddress;
}

CString CMyApp::GetErrorMessage(DWORD dwError)
{
  CString sError;

  //Lookup the error using FormatMessage
  LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26490)
  const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                                     nullptr, dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT), reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};

  if (dwReturn)
  {
    sError = lpBuffer;
    LocalFree(lpBuffer);
  }

  return sError;
}

CString CMyApp::GetIpErrorString(IP_STATUS dwError)
{
  DWORD dwBuffer{0};
  ::GetIpErrorString(dwError, nullptr, &dwBuffer);
  CStringW sUnicodeError;
  ::GetIpErrorString(dwError, sUnicodeError.GetBuffer(dwBuffer), &dwBuffer);
  sUnicodeError.ReleaseBuffer();
  return CString{sUnicodeError};
}

CString CMyApp::RTTAsString(DWORD dwRTT)
{
  CString sMsg;
  if (dwRTT == 0)
    sMsg = _T("<1ms");
  else
    sMsg.Format(_T("%ums"), dwRTT);
  return sMsg;
}

BOOL CMyApp::InitInstance()
{
  //Parse the command line
  CMyCommandLineInfo rCmdInfo;
  for (int i{1}; i<__argc; i++)
  {
  #ifdef _UNICODE
#pragma warning(suppress: 26429 26481)
    LPCTSTR pszParam{__wargv[i]};
  #else
#pragma warning(suppress: 26429 26481)
    LPCTSTR pszParam{__argv[i]};
  #endif
    BOOL bFlag{FALSE};
    const BOOL bLast{(i + 1) == __argc};
#pragma warning(suppress: 26481)
    if (pszParam[0] == _T('-') || pszParam[0] == _T('/'))
    {
      //remove flag specifier
      bFlag = TRUE;
#pragma warning(suppress: 26481)
      ++pszParam;
    }
#pragma warning(suppress: 26486)
    rCmdInfo.ParseParam(pszParam, bFlag, bLast);
  }

  //Check that the host to ping has been specified
  if (rCmdInfo.m_sHostToResolve.IsEmpty())
  {
    _tprintf(_T("Usage: cping [-t] [-a] [-n:count] [-l:size] [-i:TTL]\n"));
    _tprintf(_T("             [-w timeout] [-S:BoundAddress] [-4 | -6] destination\n\n"));
    _tprintf(_T("  Options:\n"));
    _tprintf(_T("        -t              Ping the specified host until stopped.\n"));
    _tprintf(_T("                        To stop - type Control-C.\n"));
    _tprintf(_T("        -a              Resolve addresses to hostnames.\n"));
    _tprintf(_T("        -n:count        Number of echo requests to send.\n"));
    _tprintf(_T("        -l:size         Send data buffer size.\n"));
    _tprintf(_T("        -f              Set Don't Fragment flag in packet.\n"));
    _tprintf(_T("        -i:TTL          Time To Live.\n"));
    _tprintf(_T("        -v:TOS          Type Of Service.\n"));
    _tprintf(_T("        -w:timeout      Timeout in milliseconds to wait for each reply.\n"));
    _tprintf(_T("        -4              Force using IPv6.\n"));
    _tprintf(_T("        -6              Force using IPv4.\n"));
    _tprintf(_T("        -S:BoundAddress The local address to bind to.\n"));
    return FALSE;
  }

  if (rCmdInfo.m_wDataRequestSize > 65500)
  {
    _tprintf(_T("Bad value for option -l. Valid range is 0 to 65500.\n"));
    return FALSE;
  }

  //Initialize sockets
  WSADATA wsa{};
  if (WSAStartup(MAKEWORD(1, 1), &wsa) != 0)
  {
    _tprintf(_T("Failed to initialize Winsock\n"));
    return FALSE;
  }

  bool bSuccess{false};
  const CPing p;
  CPingReplyv4 prv4;
  CPingReplyv6 prv6;
  int nRequestsSent{0};
  _tprintf(_T("Pinging %s with %u bytes of data\n"), rCmdInfo.m_sHostToResolve.GetString(), rCmdInfo.m_wDataRequestSize);
  while (true)
  {
    if (rCmdInfo.m_bIPv6)
    {
      if (rCmdInfo.m_sLocalBoundAddress.IsEmpty())
#pragma warning(suppress: 26486)
        bSuccess = p.PingUsingICMPv6(rCmdInfo.m_sHostToResolve, prv6, rCmdInfo.m_nTTL, rCmdInfo.m_dwTimeout,
                                     rCmdInfo.m_wDataRequestSize, rCmdInfo.m_nTOS, rCmdInfo.m_bDontFragment);
      else
#pragma warning(suppress: 26486)
        bSuccess = p.PingUsingICMPv6(rCmdInfo.m_sHostToResolve, prv6, rCmdInfo.m_nTTL, rCmdInfo.m_dwTimeout,
                                     rCmdInfo.m_wDataRequestSize, rCmdInfo.m_nTOS, rCmdInfo.m_bDontFragment, false, rCmdInfo.m_sLocalBoundAddress);
    }
    else
    {
      if (rCmdInfo.m_sLocalBoundAddress.IsEmpty())
#pragma warning(suppress: 26486)
        bSuccess = p.PingUsingICMPv4(rCmdInfo.m_sHostToResolve, prv4, rCmdInfo.m_nTTL, rCmdInfo.m_dwTimeout,
                                     rCmdInfo.m_wDataRequestSize, rCmdInfo.m_nTOS, rCmdInfo.m_bDontFragment);
      else
#pragma warning(suppress: 26486)
        bSuccess = p.PingUsingICMPv4(rCmdInfo.m_sHostToResolve, prv4, rCmdInfo.m_nTTL, rCmdInfo.m_dwTimeout,
                                     rCmdInfo.m_wDataRequestSize, rCmdInfo.m_nTOS, rCmdInfo.m_bDontFragment, false, rCmdInfo.m_sLocalBoundAddress);
    }

    ++nRequestsSent;
    if (bSuccess)
    {
#pragma warning(suppress: 26490)
      const SOCKADDR* pAddress{rCmdInfo.m_bIPv6 ? reinterpret_cast<SOCKADDR*>(&prv6) : reinterpret_cast<SOCKADDR*>(&prv4)};
      const int nAddressLen{rCmdInfo.m_bIPv6 ? static_cast<int>(sizeof(prv6)) : static_cast<int>(sizeof(prv4))};
      const unsigned long& nRTT{rCmdInfo.m_bIPv6 ? prv6.RTT : prv4.RTT};
      const unsigned long& nEchoReplyStatus{rCmdInfo.m_bIPv6 ? prv6.EchoReplyStatus : prv4.EchoReplyStatus};

      CString sHost;
      if (rCmdInfo.m_bResolveAddressesToHostnames)
        sHost = AddressToString(pAddress, nAddressLen, 0, nullptr);
      if (sHost.GetLength())
      {
        if (nEchoReplyStatus == IP_SUCCESS)
#pragma warning(suppress: 26472)
          _tprintf(_T("Reply from %s [%s], bytes=%d, time=%s TTL=%d\n"), AddressToString(pAddress, nAddressLen, NI_NUMERICHOST, nullptr).GetString(), sHost.GetString(), static_cast<int>(rCmdInfo.m_wDataRequestSize), RTTAsString(nRTT).GetString(), static_cast<int>(rCmdInfo.m_nTTL));
        else
          _tprintf(_T("Reply from %s [%s]: %s\n"), AddressToString(pAddress, nAddressLen, NI_NUMERICHOST, nullptr).GetString(), sHost.GetString(), GetIpErrorString(nEchoReplyStatus).GetString());
      }
      else
      {
        if (nEchoReplyStatus == IP_SUCCESS)
#pragma warning(suppress: 26472)
          _tprintf(_T("Reply from %s, bytes=%d, time=%s TTL=%d\n"), AddressToString(pAddress, nAddressLen, NI_NUMERICHOST, nullptr).GetString(), static_cast<int>(rCmdInfo.m_wDataRequestSize), RTTAsString(nRTT).GetString(), static_cast<int>(rCmdInfo.m_nTTL));
        else
          _tprintf(_T("Reply from %s: %s\n"), AddressToString(pAddress, nAddressLen, NI_NUMERICHOST, nullptr).GetString(), GetIpErrorString(nEchoReplyStatus).GetString());
      }
    }
    else
    {
      const DWORD dwError{GetLastError()};
      _tprintf(_T("%s\n"), GetErrorMessage(dwError).GetString());
    }

    //Prepare for the next loop around?
    if (!rCmdInfo.m_bPingTillStopped)
    {
      if (nRequestsSent == rCmdInfo.m_nRequestsToSend)
        break;
    }
  }

  //Cleanup usage of winsock
  WSACleanup();

  return FALSE;
}
